import json
import logging
import socket

"""
   Copyright: 2015,2016    Carl F. Karsten <carl@nextdayvideo.com>, Ryan Verner <ryan@nextdayvideo.com.au>

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:
 
   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.
"""


log = logging.getLogger('Connection')


class Connection(object):
    def __init__(self, host, port=9999):
        self.host = host
        self.port = port

        log.info('Establishing connection to %s', host)
        self.sock = socket.create_connection((host, port))

        log.debug('Connection successful \o/')
        ip = self.sock.getpeername()[0]
        log.debug('Remote-IP is %s', ip)

    def fetch_config(self):
        log.info('Reading server-config')
        fd = self.sock.makefile('rw')
        fd.write('get_config\n')
        fd.flush()

        while True:
            line = fd.readline()
            signal, _, args = line.partition(' ')

            if signal != 'server_config':
                continue

            server_config = json.loads(args)
            return server_config



#!/usr/bin/env python3
# Based upon:
# https://github.com/voc/voctomix/tree/voctopanel/example-scripts/voctolight

from enum import Enum
import argparse
import asyncio
import configparser
import json
import os.path


class Config(configparser.ConfigParser, object):
    config_files = [
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     'default-config.ini'),
        os.path.join(os.path.dirname(os.path.realpath(__file__)),
                     'config.ini'),
        '/etc/voctomix/voctolight.ini',
        '/etc/voctolight.ini',
        os.path.expanduser('~/.voctolight.ini'),
    ]

    def __init__(self, cmd_line_config=None):
        super().__init__()
        self.cmd_line_config = cmd_line_config
        self._read_config()

    def _read_config(self):
        self.read(self.config_files)
        if self.cmd_line_config:
            self.read_file(self.cmd_line_config)
            self.cmd_line_config.seek(0)

    def setup_with_server_config(self, server_config):
        self.clear()
        self._read_config()
        self.read_dict(server_config)

    def getlist(self, section, option):
        return [x.strip() for x in self.get(section, option).split(',')]


class Connection(asyncio.Protocol):
    def __init__(self, interpreter):
        self.interpreter = interpreter

    def connect(self, host, port=9999):
        self.loop = asyncio.get_event_loop()
        self.conn = self.loop.create_connection(lambda: self, host, port)
        self.loop.run_until_complete(conn.conn)

    def send(self, message):
        self.transport.write(message.encode())
        self.transport.write('\n'.encode())

    def connection_made(self, transport):
        self.transport = transport

    def data_received(self, data):
        # print('data received: {}'.format(data.decode()))
        lines = data.decode().rstrip('\n').split('\n')
        for line in lines:
            interpreter.handler(line)

    def connection_lost(self, exc):
        print('server closed the connection')
        asyncio.get_event_loop().stop()


# FIXME: Duplicate from videomix.py
class CompositeModes(Enum):
        fullscreen = 0
        side_by_side_equal = 1
        side_by_side_preview = 2
        picture_in_picture = 3


class Interpreter(object):
    a_or_b = False
    primary = False
    composite_mode = CompositeModes.fullscreen

    def __init__(self, actor, config, debug=False):
        self.config = config
        self.actor = actor
        self.debug = debug
        actor.reset_led()
        if self.debug:
            print('LED has been reset to off')

    def compute_state(self):
        if self.composite_mode == CompositeModes.fullscreen:
            return self.primary
        else:
            return self.a_or_b

    def handler(self, response):
        words = response.split()
        signal = words[0]
        args = words[1:]
        try:
            handler = getattr(self, 'handle_{}'.format(signal))
        except AttributeError:
            print('Ignoring signal', signal)
        else:
            handler(args)
            enable = interpreter.compute_state()
            actor.enable_tally(enable)
            if self.debug:
                print('LED has been set to {}.'.format(
                    'on' if enable else 'off'))

    def handle_video_status(self, cams):
        mycam = self.config.get('light', 'cam')
        if mycam in cams:
            self.a_or_b = True
        else:
            self.a_or_b = False

        self.primary = (cams[0] == mycam)
        # print ('Is primary?', self.primary)

    def handle_composite_mode(self, mode_list):
        mode = mode_list[0]
        if mode == 'fullscreen':
            self.composite_mode = CompositeModes.fullscreen
        elif mode == 'side_by_side_equal':
            self.composite_mode = CompositeModes.side_by_side_equal
        elif mode == 'side_by_side_preview':
            self.composite_mode = CompositeModes.side_by_side_preview
        elif mode == 'picture_in_picture':
            self.composite_mode = CompositeModes.picture_in_picture
        else:
            print('Cannot handle', mode, 'of type', type(mode))

    def handle_server_config(self, args):
        server_config_json = ' '.join(args)
        server_config = json.loads(server_config_json)
        self.config.setup_with_server_config(server_config)


class RPiGPIODriver:
    def __init__(self, config):
        self._import()
        self.gpios = config.get('light', 'gpios').split(',')
        self.gpio_red = int(config.get('light', 'gpio_red'))
        self.reset_led()

    def _import(self):
        global GPIO
        import RPi.GPIO as GPIO

    def reset_led(self):
        GPIO.setmode(GPIO.BOARD)
        GPIO.setwarnings(False)
        for gpio in self.gpios:
            gpio = int(gpio)
            GPIO.setup(gpio, GPIO.OUT)
            GPIO.output(gpio, GPIO.HIGH)

    def enable_tally(self, enable):
        if enable:
            GPIO.output(self.gpio_red, GPIO.LOW)
        else:
            GPIO.output(self.gpio_red, GPIO.HIGH)


class DummyDriver:
    def __init__(self, config):
        pass

    def reset_led(self):
        pass

    def enable_tally(self, enable):
        pass


class SerialDTRDriver:
    def __init__(self, config):
        self.fn = config.get('light', 'port')
        self.fd = None

    def reset_led(self):
        if self.fd:
            self.fd.close()
            self.fd = None

    def enable_tally(self, enable):
        if enable:
            self.fd = open(self.fn)
        else:
            self.reset_led()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Tallylight controlling daemon for voctomix.')
    parser.add_argument(
        '--config', type=open, help='Use a specific config file')
    parser.add_argument(
        '--debug', action='store_true',
        help='Show what would be done instead of toggling lights')
    args = parser.parse_args()
    config = Config(cmd_line_config=args.config)
    driver = config.get('light', 'driver')
    if driver == 'dummy':
        actor = DummyDriver(config)
    elif driver == 'rpi':
        actor = RPiGPIODriver(config)
    elif driver == 'serial':
        actor = SerialDTRDriver(config)
    else:
        raise Exception('Unknown driver: ' + driver)
    interpreter = Interpreter(actor, config, debug=args.debug)
    conn = Connection(interpreter)
    conn.connect(config.get('server', 'host'))
    conn.send('get_config')
    conn.send('get_composite_mode')
    conn.send('get_video')
    try:
        conn.loop.run_forever()
    except KeyboardInterrupt:
        print('Quitting (Ctrl-C)')
    actor.reset_led()



#!/usr/bin/env python3
"""
ingest.py: source client for Voctomix.

   Copyright: 2015,2016,2017 Carl F. Karsten <carl@nextdayvideo.com>,
                             Ryan Verner <ryan@nextdayvideo.com.au>

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to
   deal in the Software without restriction, including without limitation the
   rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
   sell copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.


Features:
    Retrieves audio and video-caps config from core.
      and client config
    Uses core's clock.
    Mix and match audio and video sources muxed into one stream.
    Can display video locally, including frame count and fps.
    Defaults to test audio and video sent to local core.
"""

import argparse
from pprint import pprint
import socket
import sys

import gi
gi.require_version('Gst', '1.0')
gi.require_version('GstNet', '1.0')
from gi.repository import Gst, GstNet, GObject

from lib.connection import Connection


def mk_video_src(args, videocaps):
    # make video source part of pipeline

    d = {'attribs': args.video_attribs}

    if args.monitor:
        if args.debug:
            d['monitor'] = """tee name=t ! queue !
                    videoconvert ! fpsdisplaysink sync=false
                    t. ! queue !"""
        else:
            d['monitor'] = """tee name=t ! queue !
                    videoconvert ! autovideosink sync=false
                    t. ! queue !"""
    else:
        d['monitor'] = ""

    if args.video_source == 'dv':
        video_src = """
            dv1394src name=videosrc {attribs} !
        dvdemux name=demux !
        queue max-size-time=4000000000 !
        dvdec !
                {monitor}
        deinterlace mode=1 !
        videoconvert !
        videorate !
        videoscale !
            """

    elif args.video_source == 'hdv':
        video_src = """
            hdv1394src {attribs} name=videosrc !
        tsdemux !
        queue max-size-time=4000000000 !
        decodebin !
                {monitor}
        deinterlace mode=1 !
        videorate !
        videoscale !
        videoconvert !
            """

    elif args.video_source == 'hdmi2usb':
        # https://hdmi2usb.tv
        # Note: this code works with 720p
        video_src = """
            v4l2src {attribs} name=videosrc !
                queue max-size-time=4000000000 !
        image/jpeg,width=1280,height=720 !
                jpegdec !
                {monitor}
                videoconvert !
                videoscale !
                videorate !
            """

    elif args.video_source == 'ximage':
        video_src = """
            ximagesrc {attribs} name=videosrc
                   use-damage=false !
                {monitor}
        videoconvert !
                videorate !
                videoscale !
            """
        # startx=0 starty=0 endx=1919 endy=1079 !

    elif args.video_source == 'blackmagic':
        video_src = """
            decklinkvideosrc {attribs} !
                {monitor}
                videoconvert !
                videorate !
                videoscale !
            """
        # yadif !
        # deinterlace

    elif args.video_source == 'png':
        video_src = """
            multifilesrc {attribs}
                caps="image/png" !
            pngdec !
            videoscale !
                {monitor}
            videoconvert !
            """

    elif args.video_source == 'file':
        video_src = """
            multifilesrc {attribs} !
            decodebin !
            videoscale !
                {monitor}
            videoconvert !
            """

    elif args.video_source == 'test':

        # things to render as text ontop of test video
        d['hostname'] = socket.gethostname()
        d['videocaps'] = videocaps

        video_src = """
videotestsrc name=videosrc {attribs} !
    clockoverlay
        text="Source:{hostname}\nCaps:{videocaps}\nAttribs:{attribs}\n"
        halignment=left line-alignment=left !
    {monitor}
            """

    video_src = video_src.format(**d)

    video_src += videocaps + "!\n"

    return video_src


def mk_audio_src(args, audiocaps):

    d = {
        'attribs': args.audio_attribs,
        'base_audio_attribs': 'provide-clock=false slave-method=re-timestamp'
    }

    if args.audio_source in ['dv', 'hdv']:
        # this only works if video is from DV also.
        # or some gst source that gets demux ed
        audio_src = """
            demux.audio !
                queue !
                audioconvert !
                """

    elif args.audio_source == 'pulse':
        audio_src = """
                pulsesrc {attribs} {base_audio_attribs} name=audiosrc !
                queue max-size-time=4000000000 ! audiorate !
                """

    elif args.audio_source == 'alsa':
        audio_src = """
                alsasrc {attribs} {base_audio_attribs} name=audiosrc !
                queue max-size-time=4000000000 ! audiorate !
                """

    elif args.audio_source == 'blackmagic':
        audio_src = """
            decklinkaudiosrc {attribs} !
            """

    elif args.audio_source == 'test':
        audio_src = """
            audiotestsrc {attribs} name=audiosrc freq=330 !
            """
    audio_src = audio_src.format(**d)

    audio_src += audiocaps + "!\n"

    return audio_src


def mk_client(core_ip, port):

    client = "tcpclientsink host={host} port={port}".format(
            host=core_ip, port=port)

    return client


def mk_pipeline(args, server_caps, core_ip):

    video_src = mk_video_src(args, server_caps['videocaps'])
    audio_src = mk_audio_src(args, server_caps['audiocaps'])

    client = mk_client(core_ip, args.port)

    pipeline = """
    {video_src}
     mux.
    {audio_src}
     mux.
            matroskamux name=mux !
    {client}
    """.format(video_src=video_src, audio_src=audio_src, client=client)

    # remove blank lines to make it more human readable
    while "\n\n" in pipeline:
        pipeline = pipeline.replace("\n\n", "\n")

    print(pipeline)

    if args.debug:
        gst_cmd = "gst-launch-1.0 {}".format(pipeline)

        # escape the ! because
        # asl2: ! is interpreted as a command history metacharacter
        gst_cmd = gst_cmd.replace("!", " \! ")

        # remove all the \n to make it easy to cut/paste into shell
        gst_cmd = gst_cmd.replace("\n", " ")
        while "  " in gst_cmd:
            gst_cmd = gst_cmd.replace("  ", " ")
        print("-"*78)
        print(gst_cmd)
        print("-"*78)

    return pipeline

def get_server_conf(core_ip, source_id, args):

    # establish a synchronus connection to server
    conn = Connection(core_ip)

    # fetch config from server
    server_config = conn.fetch_config()

    # Pull out the configs relevant to this client
    server_conf = {
        'videocaps': server_config['mix']['videocaps'],
        'audiocaps': server_config['mix']['audiocaps']
        }

    if source_id is not None:

        # get conf from server for this source,
        d=server_config[source_id]
        if args.debug:
            pprint(d)

        # stomp all over command line values
        # this is backwards: command line should override conf file.
        for k in d:
            if args.debug:
                print('--{}="{}"'.format(k,d[k]))
            # python argparse converts a-b to a_b, so we will to too.
            args.__setattr__(k.replace("-", "_"),d[k])

    return server_conf, args


def get_clock(core_ip, core_clock_port=9998):

    clock = GstNet.NetClientClock.new(
        'voctocore', core_ip, core_clock_port, 0)

    print('obtained NetClientClock from host: {ip}:{port}'.format(
        ip=core_ip, port=core_clock_port))

    print('waiting for NetClientClock to sync...')
    clock.wait_for_sync(Gst.CLOCK_TIME_NONE)
    print('synced with NetClientClock.')

    return clock


def run_pipeline(pipeline, clock, audio_delay=0, video_delay=0):

    def on_eos(bus, message):
        print('Received EOS-Signal')
        sys.exit(1)

    def on_error(bus, message):
        print('Received Error-Signal')
        (error, debug) = message.parse_error()
        print('Error-Details: #%u: %s' % (error.code, debug))
        sys.exit(2)

    print('starting pipeline...')
    senderPipeline = Gst.parse_launch(pipeline)
    senderPipeline.use_clock(clock)

    # Delay video/audio if required
    NS_TO_MS = 100000

    if video_delay > 0:
        print('Adjusting video sync: [{} milliseconds]'.format(video_delay))
        video_delay = video_delay * NS_TO_MS
        videosrc = senderPipeline.get_by_name('videosrc')
        videosrc.get_static_pad('src').set_offset(video_delay)

    if audio_delay > 0:
        print('Adjusting audio sync: [{} milliseconds]'.format(audio_delay))
        audio_delay = audio_delay * NS_TO_MS
        audiosrc = senderPipeline.get_by_name('audiosrc')
        audiosrc.get_static_pad('src').set_offset(audio_delay)

    # Binding End-of-Stream-Signal on Source-Pipeline
    senderPipeline.bus.add_signal_watch()
    senderPipeline.bus.connect("message::eos", on_eos)
    senderPipeline.bus.connect("message::error", on_error)

    print("playing...")
    senderPipeline.set_state(Gst.State.PLAYING)

    mainloop = GObject.MainLoop()
    try:
        mainloop.run()
    except KeyboardInterrupt:
        print('Terminated via Ctrl-C')

    print('Shutting down...')
    senderPipeline.set_state(Gst.State.NULL)
    print('Done.')

    return


def get_args():
    parser = argparse.ArgumentParser(
            description='''Vocto-ingest Client with Net-time support.
            Gst caps are retrieved from the server.
            Run without parameters: send test av to localhost:10000
            ''')

    parser.add_argument(
        '-v', '--verbose', action='count', default=0,
        help="Also print INFO and DEBUG messages.")

    parser.add_argument('--source-id', action='store',
            help="get config from server using this id.")

    parser.add_argument(
        '--video-source', action='store',
        choices=[
            'dv', 'hdv', 'hdmi2usb', 'blackmagic',
            'ximage', 'png', 'file', 'test'],
        default='test',
        help="Where to get video from")

    parser.add_argument(
        '--video-attribs', action='store', default='',
        help="misc video attributes for gst")

    parser.add_argument(
        '--video-delay', action='store',
        default=0,
        type=int,
        help="delay video by this many milliseconds")

    parser.add_argument(
        '--audio-source', action='store',
        choices=['dv', 'alsa', 'pulse', 'blackmagic', 'test'],
        default='test',
        help="Where to get audio from")

    parser.add_argument(
        '--audio-attribs', action='store',
        default='',
        help="misc audio attributes for gst")

    parser.add_argument(
        '--audio-delay', action='store',
        default=0,
        type=int,
        help="delay audio by this many milliseconds")

    parser.add_argument(
        '-m', '--monitor', action='store_true',
        help="fps display sink")

    parser.add_argument(
        '--host', action='store',
        default='localhost',
        help="hostname of vocto core")

    parser.add_argument(
        '--port', action='store',
        default='10000',
        help="port of vocto core")

    parser.add_argument(
        '--debug', action='store_true',
        help="debugging things, like dump a  gst-launch-1.0 command")

    args = parser.parse_args()

    return args


def main():
    GObject.threads_init()
    Gst.init([])

    args = get_args()
    core_ip = socket.gethostbyname(args.host)

    server_caps, args = get_server_conf(core_ip, args.source_id, args)

    pipeline = mk_pipeline(args, server_caps, core_ip)

    clock = get_clock(core_ip)

    run_pipeline(pipeline, clock, args.audio_delay, args.video_delay)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import asyncio
import json

import gbulb
import gi

gi.require_version('Gst', '1.0')
gi.require_version('GstNet', '1.0')
from gi.repository import Gst, GstNet


CONFIG = {
    'mix': {
        'videocaps': 'video/x-raw,format=I420,width=1280,height=720,'
                     'framerate=30000/1001,pixel-aspect-ratio=1/1',
        'audiocaps': 'audio/x-raw,format=S16LE,channels=2,'
                     'layout=interleaved,rate=48000',
    }
}


class VoctoMixProtocol(asyncio.Protocol):
    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        print('Control connection from {}'.format(peername))
        self.transport = transport

    def data_received(self, data):
        message = data.decode()

        if message == 'get_config\n':
            self.transport.write(
                'server_config {}\n'
                .format(json.dumps(CONFIG)).encode('utf-8')
            )
        else:
            print('Unknown control command: {!r}'.format(message))


class NetTimeClock(object):
    def __init__(self):
        clock = Gst.SystemClock().obtain()
        self.ntp = GstNet.NetTimeProvider.new(clock, '0.0.0.0', 9998)
        print('Clock listening on UDP port {}'.format(9998))

    def stop(self):
        del self.ntp  # FIXME: How is one supposed to shut it down?


class VideoSink(object):
    def __init__(self):
        print('Listening for stream on port {}'.format(10000))
        self.start()

    def start(self):
        self.pipeline = Gst.parse_launch(
            'tcpserversrc host=0.0.0.0 port=10000 name=server ! '
            'decodebin ! videoconvert ! xvimagesink'
        )
        self.pipeline.bus.add_signal_watch()
        self.pipeline.bus.connect("message::eos", self.on_eos)
        self.pipeline.set_state(Gst.State.PLAYING)

    def stop(self):
        self.pipeline.set_state(Gst.State.NULL)
        del self.pipeline

    def on_eos(self, bus, message):
        self.stop()
        self.start()


def main():
    Gst.init([])
    gbulb.install()

    clock = NetTimeClock()
    sink = VideoSink()

    loop = asyncio.get_event_loop()
    coro = loop.create_server(VoctoMixProtocol, '0.0.0.0', 9999)
    server = loop.run_until_complete(coro)

    print('Control server listening on {}'.format(
          server.sockets[0].getsockname()))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    server.close()
    clock.stop()
    sink.stop()
    try:
        loop.run_until_complete(server.wait_closed())
    except KeyboardInterrupt:
        # Gets carried through to here under glib main loop, for some reason...
        pass
    loop.close()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright: 2015,2016    Carl F. Karsten <carl@nextdayvideo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.


import argparse
import socket
import datetime
import sys


def capture_cuts(sock):
    """Listen to a voctocore control socket, and yield on cuts"""
    fd = sock.makefile('rw')
    for line in fd:
        words = line.rstrip('\n').split(' ')

        signal = words[0]
        args = words[1:]

        if signal == 'message' and args[0] == 'cut':
            yield


def main():
    p = argparse.ArgumentParser()
    p.add_argument('--host', default='localhost',
                   help='Hostname of voctocore')
    p.add_argument('--port', type=int, default=9999,
                   help='Port to connect to, on voctocore')
    p.add_argument('--file', type=argparse.FileType('a'),
                   help='Filename to write cuts to')
    args = p.parse_args()

    sock = socket.create_connection((args.host, args.port))
    for cut in capture_cuts(sock):
        ts = datetime.datetime.now().strftime("%Y-%m-%d/%H_%M_%S")
        print(ts)
        sys.stdout.flush()
        if args.file:
            args.file.write('%s\n' % ts)
            args.file.flush()


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
